// ***********************************************
//    SOUBOR:        Vpis 13.13
//
//    EL:          Ukzat propojen seznam
//    POZNMKY:
//
//  COPYRIGHT:  Copyright (C) 1998 Liberty Associates, Inc.
//                All Rights Reserved
//
// Ukazuje objektov orientovan pstup k propojenm
// seznamm. Seznam deleguje innost uzlu. Uzel je
// abstraktn datov typ. Pouvaj se ti typy uzl:
// uzly hlavy, uzly ocasu a vnitn uzly. Data
// obsahuj pouze vnitn uzly.
//
// Tda Data je vytvoena, aby slouila jako objekt
// k uchovn v danm propojenm seznamu.
//
// ***********************************************


#include <iostream>
using namespace std;

enum { kJeMensi, kJeVetsi, kJeStejny};

// Tda Data pro vloen propojenho seznamu. Kad 
// tda v propojenm seznamu mus podporovat dv metody:
// Zobrazit (zobraz hodnotu) a
// Porovnat (vrt relativn pozici)
class Data
{
public:
    Data(int hodnota):mojeHodnota(hodnota){}
    ~Data(){}
    int Porovnat(const Data &);
    void Zobrazit() { cout << mojeHodnota << endl; }
private:
    int mojeHodnota;
};

// Porovnat se pouv k uren, kam do seznamu
// urit objekt pat.
int Data::Porovnat(const Data & jinaData)
{
    if (mojeHodnota < jinaData.mojeHodnota)
        return kJeMensi;
    if (mojeHodnota > jinaData.mojeHodnota)
        return kJeVetsi;
    else
        return kJeStejny;
}

// Dopedn deklarace
class Uzel;
class UzelHlavy;
class UzelOcasu;
class VnitrniUzel;

// ADT pedstavujc objekt uzlu v seznamu
// Kad odvozen tda mus pekrt Vlozit a Zobrazit
class Uzel
{
public:
    Uzel(){}
    virtual ~Uzel(){}
    virtual Uzel * Vlozit(Data * taData)=0;
    virtual void Zobrazit() = 0;
private:
};

// Toto je uzel, kter obsahuje vlastn objekt
// V tomto ppad jde o objekt typu Data
// Uvidme, jak to uinit obecnji, a si budeme
// povdat o ablonch
class VnitrniUzel: public Uzel
{
public:
    VnitrniUzel(Data * taData, Uzel * dalsi);
    ~VnitrniUzel(){ delete mujDalsi; delete mojeData; }
    virtual Uzel * Vlozit(Data * taData);
    // Delegovat!
    virtual void Zobrazit() { mojeData->Zobrazit(); mujDalsi->Zobrazit();}  

private:
    Data * mojeData;    // samotn data
    Uzel * mujDalsi;    // ukazuje na dal uzel v propojenm seznamu
};

// Konstruktor jenom inicializuje
VnitrniUzel::VnitrniUzel(Data * taData, Uzel * dalsi):
mojeData(taData),mujDalsi(dalsi)
{
}

// Zklad celho seznamu
// Kdy do seznamu vlote nov objekt, 
// ped se uzlu, kter ur jeho pozici
// a vlo jej do seznamu
Uzel * VnitrniUzel::Vlozit(Data * taData)
{

    // Je ten nov hoch vt nebo men ne j?
    int vysledek = mojeData->Porovnat(*taData);


    switch(vysledek)
    {
    // Je li stejn jako j, pak podle konvence pat pede m
    case kJeStejny:        // projt dl
    case kJeVetsi:         // nov data pat pede m
        {
            VnitrniUzel * datovyUzel = new VnitrniUzel(taData, this);
            return datovyUzel;
        }

    // Je vt ne j, take jej pedme dalmu
    // uzlu a ponechme na NM, a se star.
    case kJeMensi:
        mujDalsi = mujDalsi->Vlozit(taData);
        return this;
    }
    return this;  // naplnit MSC
}


// Uzel ocasu je jen hldka

class UzelOcasu : public Uzel
{
public:
    UzelOcasu(){}
    ~UzelOcasu(){}
    virtual Uzel * Vlozit(Data * taData);
    virtual void Zobrazit() { }

private:

};

// Pokud njak data pijdou a ke mn, mus se vloit
// pede m, protoe j jsem ocas a za mnou u nen NIC.
Uzel * UzelOcasu::Vlozit(Data * taData)
{
    VnitrniUzel * datovyUzel = new VnitrniUzel(taData, this);
    return datovyUzel;
}

// Uzel hlavy neobsahuje dn data, jenom 
// ukazuje na pln zatek seznamu
class UzelHlavy : public Uzel
{
public:
    UzelHlavy();
    ~UzelHlavy() { delete mujDalsi; }
    virtual Uzel * Vlozit(Data * taData);
    virtual void Zobrazit() { mujDalsi->Zobrazit(); }
private:
    Uzel * mujDalsi;
};

// Jakmile se vytvo hlava, vytvo se
// tak ocas.
UzelHlavy::UzelHlavy()
{
    mujDalsi = new UzelOcasu;
}

// Ped hlavou nic nen, take data
// jen pedme dalmu uzlu.
Uzel * UzelHlavy::Vlozit(Data * taData)
{
    mujDalsi = mujDalsi->Vlozit(taData);
    return this;
}

// Mm vechny zsluhy a pitom nic nedlm.
class PropojenySeznam
{
public:
    PropojenySeznam();
    ~PropojenySeznam() { delete mojeHlava; }
    void Vlozit(Data * taData);
    void ZobrazitVse() { mojeHlava->Zobrazit(); }
private:
    UzelHlavy * mojeHlava;
};

// Pi zrozen vytvom uzel hlavy.
// Ten vytvo uzel ocasu, take
// przdn seznam ukazuje na hlavu, kter
// ukazuje na ocas a mezi nimi nic nen.
PropojenySeznam::PropojenySeznam()
{
    mojeHlava = new UzelHlavy;
}

// Deleguji, deleguje, delegujeme
void PropojenySeznam::Vlozit(Data * pData)
{
    mojeHlava->Vlozit(pData);
}

// Testovac program
int main()
{
    Data * pData;
    int hodnota;
    PropojenySeznam ps;

    // Podat uivatele a zadn njakch hodnot
    // a vloit je do seznamu.
    for (;;)
    {
        cout << "Jaka hodnota? (konec = 0): ";
        cin >> hodnota;
        if (!hodnota)
            break;
        pData = new Data(hodnota);
        ps.Vlozit(pData);
    }

    // Nyn projt seznamem a zobrazit data
    ps.ZobrazitVse();
    return 0;  // ps vypadne z oboru platnosti a je zruen!
}
