//Vpis 16.8 - Ptelsk opertory

#include <iostream>
#include <string.h>
using namespace std;

// Zkladn etzcov tda
class Retezec
{
public:
    // Konstruktory
    Retezec();
    Retezec(const char *const);
    Retezec(const Retezec &);
    ~Retezec();

    // Peten opertory
    char & operator[](int odsazeni);
    char operator[](int odsazeni) const;
    Retezec operator+(const Retezec&);
    friend Retezec operator+(const Retezec&, const Retezec&);
    void operator+=(const Retezec&);
    Retezec & operator= (const Retezec &);

    // Obecn pstupov funkce
    int ZjistitDelku()const { return jehoDelka; }
    const char * PrevzitRetezec() const { return jehoRetezec; }

private:
    Retezec (int);         // Soukrom konstruktor
    char * jehoRetezec;
    unsigned short jehoDelka;
};

// Vchoz konstruktor vytv etzec dlky 0 bajt
Retezec::Retezec()
{
    jehoRetezec = new char[1];
    jehoRetezec[0] = '\0';
    jehoDelka=0;
    // cout << "\tVchoz konstruktor Retezec\n";
    // PocetKonstruktoru++;
}

// Soukrom (pomocn) konstruktor pouvan pouze
// metodami tto tdy k vytvoen novho etzce
// poadovan velikosti.  Naplnn nulami.
Retezec::Retezec(int delka)
{
    jehoRetezec = new char[delka+1];
    for (int i = 0; i<=delka; i++)
        jehoRetezec[i] = '\0';
    jehoDelka=delka;
    // cout << "\tKonstruktor Retezec(int)\n";
    // PocetKonstruktoru++;
}

// Pevd pole znak na Retezec
Retezec::Retezec(const char * const cRetezec)
{
    jehoDelka = strlen(cRetezec);
    jehoRetezec = new char[jehoDelka+1];
    for (int i = 0; i<jehoDelka; i++)
        jehoRetezec[i] = cRetezec[i];
    jehoRetezec[jehoDelka]='\0';
    // cout << "\tKonstruktor Retezec(char*)\n";
    // PocetKonstruktoru++;
}

// Koprovac konstruktor
Retezec::Retezec (const Retezec & pravaStrana)
{
    jehoDelka=pravaStrana.ZjistitDelku();
    jehoRetezec = new char[jehoDelka+1];
    for (int i = 0; i<jehoDelka;i++)
        jehoRetezec[i] = pravaStrana[i];
    jehoRetezec[jehoDelka] = '\0';
    // cout << "\tKonstruktor Retezec(Retezec&)\n";
    // PocetKonstruktoru++;
}

// Destruktor, uvoluje alokovanou pam
Retezec::~Retezec ()
{
    delete [] jehoRetezec;
    jehoDelka = 0;
    // cout << "\tDestruktor Retezec\n";
}

// Opertor rovnosti, uvoluje existujc pam
// a pak kopruje etzec a velikost
Retezec& Retezec::operator=(const Retezec & pravaStrana)
{
    if (this == &pravaStrana)
        return *this;
    delete [] jehoRetezec;
    jehoDelka=pravaStrana.ZjistitDelku();
    jehoRetezec = new char[jehoDelka+1];
    for (int i = 0; i<jehoDelka;i++)
        jehoRetezec[i] = pravaStrana[i];
    jehoRetezec[jehoDelka] = '\0';
    return *this;
    // cout << "\tOpertor rovnosti Retezec\n";
}

// Nekonstantn opertor odsazen, vrac
// odkaz na znak, aby jej bylo mon
// zmnit!
char & Retezec::operator[](int odsazeni)
{
    if (odsazeni > jehoDelka)
        return jehoRetezec[jehoDelka-1];
    else
        return jehoRetezec[odsazeni];
}

// Konstantn opertor odsazen pro pouit 
// s konstantnmi objekty (viz koprovac konstruktor!)
char Retezec::operator[](int odsazeni) const
{
    if (odsazeni > jehoDelka)
        return jehoRetezec[jehoDelka-1];
    else
        return jehoRetezec[odsazeni];
}

// Vytv nov etzec pidnm aktulnho
// etzce k prav stran
Retezec Retezec::operator+(const Retezec& pravaStrana)
{
    int  celkovaDelka = jehoDelka + pravaStrana.ZjistitDelku();
    Retezec docasny(celkovaDelka);
    int i, j;
    for (i = 0; i<jehoDelka; i++)
        docasny[i] = jehoRetezec[i];
    for (j = 0, i = jehoDelka; j<pravaStrana.ZjistitDelku(); j++, i++)
        docasny[i] = pravaStrana[j];
    docasny[celkovaDelka]='\0';
    return docasny;
}

// Vytv nov etzec pidnm
// jednoho etzce k druhmu
Retezec operator+(const Retezec& levaStrana, const Retezec& pravaStrana)
{
    int  celkovaDelka = levaStrana.ZjistitDelku() + 
                        pravaStrana.ZjistitDelku();
    Retezec docasny(celkovaDelka);
    int i, j;
    for (i = 0; i<levaStrana.ZjistitDelku(); i++)
        docasny[i] = levaStrana[i];
    for (j = 0, i = levaStrana.ZjistitDelku(); 
                    j<pravaStrana.ZjistitDelku(); j++, i++)
        docasny[i] = pravaStrana[j];
    docasny[celkovaDelka]='\0';
    return docasny;
}

int main()
{
    Retezec s1("Retezec jedna ");
    Retezec s2("Retezec dva ");
    char *c1 = { "Z-Retezec jedna " } ;
    Retezec s3;
    Retezec s4;
    Retezec s5;

    cout << "s1: " << s1.PrevzitRetezec() << endl;
    cout << "s2: " << s2.PrevzitRetezec() << endl;
    cout << "c1: " << c1 << endl;
    s3 = s1 + s2;
    cout << "s3: " << s3.PrevzitRetezec() << endl;
    s4 = s1 + c1;
    cout << "s4: " << s4.PrevzitRetezec() << endl;
    s5 = c1 + s2;
    cout << "s5: " << s5.PrevzitRetezec() << endl;
    return 0;
}
