#define DEBUG
#define UKAZ_INVARIANTY
#include <iostream>
#include <string.h>
using namespace std;

#ifndef DEBUG
	#define ASSERT(x)
#else
	#define ASSERT(x) \
			if (! (x)) \
			{ \
				cout << "CHYBA!! Makro assert " << #x << " neproslo\n"; \
				cout << " na radku " << __LINE__  << "\n"; \
				cout << " v souboru " << __FILE__ << "\n";  \
			}
#endif


const int FALSE = 0;
const int TRUE = 1;
typedef int BOOL;


class Retezec
{
public:
	// konstruktory
	Retezec();
	Retezec(const char *const);
	Retezec(const Retezec &);
	~Retezec();

	char & operator[](int odskok);
	char operator[](int odskok) const;

	Retezec & operator= (const Retezec &);
	int ZiskejDelku()const { return delkaRetezce; }
	const char * ZiskejRetezec() const { return retezec; }
	BOOL Invarianty() const;

private:
	Retezec (int);         // privtn konstruktor
	char * retezec;
	// unsigned short delkaRetezce;
	int delkaRetezce;
};

// vchoz konstruktor vytvo etzec o 0 bajtech
Retezec::Retezec()
{
	retezec = new char[1];
	retezec[0] = '\0';
	delkaRetezce=0;
	ASSERT(Invarianty());
}

// Privtn (pomocn) konstruktor, pouvan pouze
// metodami tdy pro vytvoen novho etzce poadovan
// dlky. etzec je zaplnn nulami.
Retezec::Retezec(int delka)
{
	retezec = new char[delka+1];
	for (int i = 0; i<=delka; i++)
		retezec[i] = '\0';
	delkaRetezce=delka;
	ASSERT(Invarianty());
}

// Pevd pole etzc na tdu Retezec
Retezec::Retezec(const char * const cRetezec)
{
	delkaRetezce = strlen(cRetezec);
	retezec = new char[delkaRetezce+1];
	for (int i = 0; i<delkaRetezce; i++)
		retezec[i] = cRetezec[i];
	retezec[delkaRetezce]='\0';
	ASSERT(Invarianty());
}

// Konstrutor pro koprovn
Retezec::Retezec (const Retezec & ptr)
{
	delkaRetezce=ptr.ZiskejDelku();
	retezec = new char[delkaRetezce+1];
	for (int i = 0; i<delkaRetezce;i++)
		retezec[i] = ptr[i];
	retezec[delkaRetezce] = '\0';
	ASSERT(Invarianty());
}

// Destruktor uvoluje zabranou pam
Retezec::~Retezec ()
{
	ASSERT(Invarianty());
	delete [] retezec;
	delkaRetezce = 0;
}

// Opertor piazen "=", uvoln existujc pam
// a zkopruje etzec a jeho velikost
Retezec& Retezec::operator=(const Retezec & ptr)
{
	ASSERT(Invarianty());
	if (this == &ptr)
		return *this;
	delete [] retezec;
	delkaRetezce=ptr.ZiskejDelku();
	retezec = new char[delkaRetezce+1];
	for (int i = 0; i<delkaRetezce;i++)
		retezec[i] = ptr[i];
	retezec[delkaRetezce] = '\0';
	ASSERT(Invarianty());
	return *this;
}

// nekonstatn opertor pro "odskok"
char & Retezec::operator[](int odskok)
{
	ASSERT(Invarianty());
	if (odskok > delkaRetezce)
	{
		ASSERT(Invarianty());
		return retezec[delkaRetezce-1];
	}
	else
	{
		ASSERT(Invarianty());
		return retezec[odskok];
	}
}

// konstatn opertor pro "odskok"
char Retezec::operator[](int odskok) const
{
	ASSERT(Invarianty());
	char retVal;
	if (odskok > delkaRetezce)
		retVal = retezec[delkaRetezce-1];
	else
		retVal = retezec[odskok];
	ASSERT(Invarianty());
	return retVal;
}

BOOL Retezec::Invarianty() const
{
	#ifdef UKAZ_INVARIANTY
		cout << "Test retezce ";
	#endif
	return ( (delkaRetezce && retezec) || (!delkaRetezce && !retezec) );
}

class Zvire
{
public:
	Zvire():vekZvirete(1),jmenoZvirete("Frantisek Zvire")
		{ASSERT(Invarianty());}
	Zvire(int, const Retezec&);
	~Zvire(){}
	int ZiskejVek() {  ASSERT(Invarianty()); return vekZvirete;}
	void NastavVek(int vek)
	{
		ASSERT(Invarianty());
		vekZvirete = vek;
		ASSERT(Invarianty());
	}
	Retezec& ZiskejJmeno()
	{
		ASSERT(Invarianty());
		return jmenoZvirete;
	}
	void NastavJmeno(const Retezec& jmeno)
	{
		ASSERT(Invarianty());
		jmenoZvirete = jmeno;
		ASSERT(Invarianty());
	}
	BOOL Invarianty();
private:
	int vekZvirete;
	Retezec jmenoZvirete;
};

Zvire::Zvire(int vek, const Retezec& jmeno):
	vekZvirete(vek),
	jmenoZvirete(jmeno)
{
	ASSERT(Invarianty());
}

BOOL Zvire::Invarianty()
{
	#ifdef UKAZ_INVARIANTY
		cout << "Test zvirete ";
	#endif
		return (vekZvirete > 0 && jmenoZvirete.ZiskejDelku());
}

int main()
{
	Zvire alik(5,"Alik");
	cout << "\n" << alik.ZiskejJmeno().ZiskejRetezec() << " ma ";
	cout << alik.ZiskejVek() << " roku.";
	alik.NastavVek(8);
	cout << "\n" << alik.ZiskejJmeno().ZiskejRetezec() << " ma ";
	cout << alik.ZiskejVek() << " roku.";
	return 0;
}