// *********************************************************************
//
// Nzev:	Tden 3. - opakovn
//
// Soubor:	ProgW301
//
// Popis:	Obsahuje ablonu pro spojov seznam a
//			ukzkov program s oetenm vjimek
//
// Tdy:	Soucastka -	Udruje sla soustek a ppadn
//						dal informace o soustkch. Je to
//						pklad tdy, kter se d pout
//						v seznamu. Vimnte si opertoru << pro
//						vytisknut informac o soustce
//						na zklad jejho typu.
//			Uzel -		Hraje roli uzlu v Seznamu
//			Seznam -	Seznam zaloen na ablon, kter
//						implementuje spojov seznam
//
//
// Autor:	Jesse Liberty (jl)
//
// Platforma:	Pentium 200 Pro. 128MB RAM MVC 5.0
//
// Clov platforma:	Nezvisl na platform
//
// Historie:	09.1994 - Prvn verze (jl)
//				04.1997 - pravy (jl)
// *********************************************************************

#include <iostream>
using namespace std;

// tdy vjimek
class Vyjimka {};
class NedostatekPameti : public Vyjimka{};
class NulovyUzel :       public Vyjimka{};
class PrazdnySeznam :    public Vyjimka{};
class ChybaMezi :        public Vyjimka{};


// **************** Soucastka ****************
// Abstraktn tda pro soustky
class Soucastka
{
public:
	Soucastka():cisloObjektuSoucastky(1) {}
	Soucastka(int cisloObjektu):cisloObjektuSoucastky(cisloObjektu){}
	virtual ~Soucastka(){};
	int ZiskejCisloObjektu() const { return cisloObjektuSoucastky; }
	virtual void Zobrazeni() const =0; // mus bt pepsno

private:
	int cisloObjektuSoucastky;
};

// Implementace ist virtuln funkce, na kterou mohou navzat
// odvozen tdy
void Soucastka::Zobrazeni() const
{
	cout << "\nSoucastka cislo: " << cisloObjektuSoucastky << endl;
}

// Tento opertor << je voln na vechny objekty soustek.
// Nemus mt atribut "friend" protoe nepistupuje k privtnm
// datm. Vol funkci Zobrazeni(), kter je polymorfn.
// Radji bychom to pepsali podle skutenho typu soustky,
// ale C++ tuto monost nedv, nepodporuje toti kontravarianci
ostream& operator<<( ostream& proud, Soucastka& pSoucastka)
{
	pSoucastka.Zobrazeni(); // virtuln kontravariance!
	return proud;
}

// **************** Soucastka Automobilu ****************
class SoucastkaAuta : public Soucastka
{
public:
	SoucastkaAuta():rokVyrobyAuta(1994){}
	SoucastkaAuta(int rok, int cisloSoucastky);
	int ZiskejRokVyroby() const { return rokVyrobyAuta; }
	virtual void Zobrazeni() const;
private:
	int rokVyrobyAuta;
};

SoucastkaAuta::SoucastkaAuta(int rok, int cisloSoucastky):
    rokVyrobyAuta(rok),
    Soucastka(cisloSoucastky)
{}

void SoucastkaAuta::Zobrazeni() const
{
	Soucastka::Zobrazeni();
	cout << "Rok vyroby modelu: " << rokVyrobyAuta << endl;
}

// **************** Soucastka Letadla ****************
class SoucastkaLetadla : public Soucastka
{
public:
	SoucastkaLetadla():cisloMotoruLetadla(1){}
	SoucastkaLetadla(int cisloMotoru, int cisloSoucastky);
	virtual void Zobrazeni() const;
	int ZiskejCisloMotoru() const { return cisloMotoruLetadla; }
private:
	int cisloMotoruLetadla;
};

SoucastkaLetadla::SoucastkaLetadla(int cisloMotoru, int cisloSoucastky):
    cisloMotoruLetadla(cisloMotoru),
	Soucastka(cisloSoucastky)
{}

void SoucastkaLetadla::Zobrazeni() const
{
	Soucastka::Zobrazeni();
	cout << "Cislo motoru: " << cisloMotoruLetadla << endl;
}

// asn deklarace tdy Seznam
template <class T>
class Seznam;

// ************** Uzel **************
// Obecn uzel pro pidn do seznamu
// **********************************

template <class T>
class Uzel
{
public:
	friend class Seznam<T>;
	Uzel (T*);
	~Uzel();
	void NastavDalsi(Uzel * uzel) { dalsiUzel = uzel; }
	Uzel * ZiskejDalsi() const;
	T * ZiskejObjekt() const;
private:
	T* objektUzlu;
	Uzel * dalsiUzel;
};

// Implementace uzlu

template <class T>
Uzel<T>::Uzel(T* pObjekt):
    objektUzlu(pObjekt),
	dalsiUzel(0)
{}

template <class T>
Uzel<T>::~Uzel()
{
	delete objektUzlu;
	objektUzlu = 0;
	delete dalsiUzel;
	dalsiUzel = 0;
}

// Vrac NULL, pokud dal uzel neexistuje
template <class T>
Uzel<T> * Uzel<T>::ZiskejDalsi() const
{
	return dalsiUzel;
}

template <class T>
T * Uzel<T>::ZiskejObjekt() const
{
	if (objektUzlu)
			return objektUzlu;
	else
		throw NulovyUzel();
}

// ************ Seznam ************
// Obecn ablona seznamu
// Zpracuje libovoln poet objekt
// ********************************
template <class T>
class Seznam
{
public:
	Seznam();
	~Seznam();

	T* Najdi(int & pozice, int cisloObjektu) const;
	T* ZiskejPrvni() const;
	void Vloz(T *);
	T* operator[](int) const;
	int ZiskejPocet() const { return pocetUzlu; }
private:
	Uzel<T> * pPrvni;
	int pocetUzlu;
};

// Implementace seznamu
template <class T>
Seznam<T>::Seznam():
    pPrvni(0),
	pocetUzlu(0)
{}

template <class T>
Seznam<T>::~Seznam()
{
	delete pPrvni;
}

template <class T>
T* Seznam<T>::ZiskejPrvni() const
{
	if (pPrvni)
		return pPrvni->objektUzlu;
	else
		throw PrazdnySeznam();
}

template <class T>
T * Seznam<T>::operator[](int odskok) const
{
	Uzel<T>* pUzel = pPrvni;

	if (!pPrvni)
		throw PrazdnySeznam();

	if (odskok > pocetUzlu)
		throw ChybaMezi();

	for (int i=0; i<odskok; i++)
		pUzel = pUzel->dalsiUzel;

	return pUzel->objektUzlu;
}

// Najde dan objekt v seznamu na zklad jednoznanho identifiktoru
template <class T>
T* Seznam<T>::Najdi(int & pozice, int cisloObjektu) const
{
	Seznam<T> * pUzel = 0;
	for (pUzel = pPrvni, pozice = 0;
	     pUzel != NULL;
		 pUzel = pUzel->dalsiUzel, pozice++ )
	{
		if (pUzel->objektUzlu->ZiskejCisloObjektu() == cisloObjektu)
			break;
	}
	if (pUzel == NULL)
		return NULL;
	else
		return pUzel->objektUzlu;
}

// Vlo do seznamu objekt, pokud je ovem uniktn
template <class T>
void Seznam<T>::Vloz(T* pObjekt)
{
	Uzel<T> * pUzel = new Uzel<T>(pObjekt);
	Uzel<T> * pAktualni = pPrvni;
	Uzel<T> * pDalsi = 0;

	int novy = pObjekt->ZiskejCisloObjektu();
	int dalsi = 0;
	pocetUzlu++;

	if (!pPrvni)
	{
		pPrvni = pUzel;
		return;
	}

	// Pokud m nov objekt men slo ne prvn,
	// stv se nov objekt prvnm v seznamu
	if (pPrvni->objektUzlu->ZiskejCisloObjektu() > novy)
	{
		pUzel->dalsiUzel = pPrvni;
		pPrvni = pUzel;
		return;
	}

	for (;;)
	{
		// Pokud nen dn dal uzel, pid se nov na konec
		if (!pAktualni->dalsiUzel)
		{
			pAktualni->dalsiUzel = pUzel;
			return;
		}

		// Pokud se nov uzel umst mezi aktuln a dal,
		// vlome ho sem, jinak pokrauje prochzen seznamu
		pDalsi = pAktualni->dalsiUzel;
		dalsi = pDalsi->objektUzlu->ZiskejCisloObjektu();
		if (dalsi > novy)
		{
			pAktualni->dalsiUzel = pUzel;
			pUzel->dalsiUzel = pDalsi;
			return;
		}
		pAktualni = pDalsi;
	}
}


int main()
{
	Seznam<Soucastka> sSeznam;
	int volba;
	int cisloObjektu;
	int hodnota;
	Soucastka * pSoucastka;
	while (1)
	{
		cout << "(0)Konec (1)Auto (2)Letadlo: ";
		cin >> volba;

		if (!volba)
			break;

		cout << "Nove cislo soucastky?: ";
		cin >> cisloObjektu;

		if (volba == 1)
		{
			cout << "Rok vyroby modelu?: ";
			cin >> hodnota;
			try
			{
				pSoucastka = new SoucastkaAuta(hodnota,cisloObjektu);
			}
			catch (NedostatekPameti)
			{
				cout << "Nedostatek pameti; Konec..." << endl;
				return 1;
			}
		}
		else
		{
			cout << "Cislo motoru?: ";
			cin >> hodnota;
			try
			{
				pSoucastka = new SoucastkaLetadla(hodnota,cisloObjektu);
			}
			catch (NedostatekPameti)
			{
				cout << "Nedostatek pameti; Konec..." << endl;
				return 1;
			}
		}
		try
		{
			sSeznam.Vloz(pSoucastka);
		}
		catch (NulovyUzel)
		{
			cout << "Seznam je narusen, uzel je nulovy!" << endl;
			return 1;
		}
		catch (PrazdnySeznam)
		{
			cout << "Seznam je prazdny!" << endl;
			return 1;
		}
	}
	try
	{
		for (int i = 0; i < sSeznam.ZiskejPocet(); i++)
			cout << *(sSeznam[i]);
	}
		catch (NulovyUzel)
		{
			cout << "Seznam je narusen, uzel je nulovy!" << endl;
			return 1;
		}
		catch (PrazdnySeznam)
		{
			cout << "Seznam je prazdny!" << endl;
			return 1;
		}
		catch (ChybaMezi)
		{
			cout << "Pokus o cteni za mezemi seznamu!" << endl;
			return 1;
		}
	return 0;
}